import math
import os
from typing import Any, Optional

from locust import HttpUser, LoadTestShape, TaskSet, between, task


class UserTasks(TaskSet):
    @task
    def get_root(self) -> None:
        endpoint = os.getenv("RELIABLY_LOCUST_ENDPOINT")
        headers = {}

        bearer_token = os.getenv("RELIABLY_LOCUST_ENDPOINT_TOKEN", "").strip()
        if bearer_token:
            headers["Authorization"] = f"Bearer {bearer_token}"

        self.client.get(endpoint, headers=headers)


class WebsiteUser(HttpUser):
    wait_time = between(1, 10)
    tasks = [UserTasks]


class StepLoadShape(LoadTestShape):
    step_time = int(os.getenv("RELIABLY_LOCUST_STEP_TIME", 1))
    step_load = int(os.getenv("RELIABLY_LOCUST_STEP_LOAD", 1))
    spawn_rate = int(os.getenv("RELIABLY_LOCUST_SPAWN_RATE", 1))
    time_limit = int(os.getenv("RELIABLY_LOCUST_TIME_LIMIT", 5))

    def tick(self) -> Optional[Any]:
        run_time = self.get_run_time()

        if run_time > self.time_limit:
            return None

        current_step = math.floor(run_time / self.step_time) + 1
        return (current_step * self.step_load, self.spawn_rate)



import os

from chaoslib import substitute
from chaosreliably.controls.autopause import configure_control


def test_autopause_method_actions():
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": [
            {
                "name": "A",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "B",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "C",
                "type": "probe",
                "provider": {"module": "os"}
            },
            {
                "name": "D",
                "type": "action",
                "provider": {"module": "os"}
            }
        ]
    }

    configure_control(experiment, {
        "method": {
            "actions": {
                "enabled": True,
                "pause_duration": 5
            }
        }
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "D"):
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 5
        if activity["name"] == "C":
            assert activities[index+1]["provider"]["module"] != "chaosreliably.activities.pauses"


def test_autopause_method_probes():
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": [
            {
                "name": "A",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "B",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "C",
                "type": "probe",
                "provider": {"module": "os"}
            },
            {
                "name": "D",
                "type": "action",
                "provider": {"module": "os"}
            }
        ]
    }

    configure_control(experiment, {
        "method": {
            "probes": {
                "enabled": True,
                "pause_duration": 5
            }
        }
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "D"):
            assert activities[index+1]["provider"]["module"] != "chaosreliably.activities.pauses"
        if activity["name"] == "C":
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 5


def test_autopause_method_probes():
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": [
            {
                "name": "A",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "B",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "C",
                "type": "probe",
                "provider": {"module": "os"}
            },
            {
                "name": "D",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "E",
                "type": "probe",
                "provider": {"module": "os"}
            },
        ]
    }

    configure_control(experiment, {
        "method": {
            "actions": {
                "enabled": False,
            },
            "probes": {
                "enabled": True,
                "pause_duration": 5
            }
        }
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "D"):
            assert activities[index+1]["provider"]["module"] != "chaosreliably.activities.pauses"
        if activity["name"] in ("C", "E"):
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 5


def test_autopause_rollbacks():
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "rollbacks": [
            {
                "name": "A",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "B",
                "type": "action",
                "provider": {"module": "os"}
            },
            {
                "name": "C",
                "type": "action",
                "provider": {"module": "os"}
            },
        ]
    }

    configure_control(experiment, {
        "rollbacks": {
            "enabled": True,
            "pause_duration": 5
        }
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "C"):
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 5


def test_autopause_ssh():
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "steady-state-hypothesis": {
            "probes":[
                {
                    "name": "A",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
                {
                    "name": "B",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
                {
                    "name": "C",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
            ]
        }
    }

    configure_control(experiment, {
        "steady-state-hypothesis": {
            "enabled": True,
            "pause_duration": 5
        }
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "C"):
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 5


def test_autopause_from_env():
    os.putenv("RELIABLY_PROBE_PAUSE_DURATION", "10.4")
    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "steady-state-hypothesis": {
            "probes":[
                {
                    "name": "A",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
                {
                    "name": "B",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
                {
                    "name": "C",
                    "type": "probe",
                    "provider": {"module": "os"}
                },
            ]
        }
    }

    configure_control(experiment, {
        "method": {
            "probes": {
                "enabled": True,
                "pause_duration": substitute({
                    "env": "type",
                    "key": "RELIABLY_PROBE_PAUSE_DURATION",
                }, {}, {})
            }
        },
    })

    activities = experiment.get("method", [])
    for index, activity in enumerate(activities):
        if activity["name"] in ("A", "B", "C"):
            assert activities[index+1]["provider"]["module"] == "chaosreliably.activities.pauses"
            assert activities[index+1]["provider"]["arguments"]["duration"] == 10.4



from typing import Optional

import httpx
from chaoslib.types import Configuration, Secrets
from logzero import logger

__all__ = ["call_endpoint"]


def call_endpoint(
    url: str,
    auth: Optional[str] = None,
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> bool:
    headers = {}
    if auth:
        headers["Authorization"] = auth

    r = httpx.get(url, headers=headers)
    if r.status_code != 200:
        logger.critical("Safeguard endpoint returned a non 200 response")
        return False

    result = r.json()
    if not result.get("ok"):
        logger.critical(
            f"Safeguard endpoint returned with an error: {result['error']}"
        )
        return False

    return True



import os
from datetime import datetime
from typing import List, Optional
from urllib.parse import urlparse

import httpx
from chaoslib.exceptions import ActivityFailed, InvalidActivity
from chaoslib.types import Configuration, Secrets
from logzero import logger

from chaosreliably import parse_duration

__all__ = ["closed_pr_ratio", "pr_duration"]


def closed_pr_ratio(
    repo: str,
    base: str = "main",
    only_opened_and_closed_during_window: bool = True,
    window: str = "5d",
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> float:
    """
    Computes a ratio of closed PRs during the given `window` in a `repo`.

    By default, only computes the ratio for PRs that were opened and closed
    during the given period. When `only_opened_and_closed_during_window` is
    not set, this computes the ratio for closed PRs in the period against
    all still opened PRs, whether they were opened before the period started
    or not.

    The former is a measure of latency for teams while the latter is more
    the throughput of the team.

    The `repo` should be given as `owner/repo` and the window should be given
    as a pattern like this: `<int>s|m|d|w` (seconds, minutes, days, weeks).
    """
    secrets = secrets or {}
    gh_token = secrets.get("github", {}).get("token")
    gh_token = os.getenv("GITHUB_TOKEN", gh_token)

    if not gh_token:
        raise InvalidActivity(
            "`closed_pr_rate` requires a github token as a secret or via the "
            "GITHUB_TOKEN environment variable"
        )

    duration = parse_duration(window)
    today = datetime.today()
    start_period = today - duration

    total_opened = 0
    total_closed_during_period = 0
    total_opened_during_period = 0

    p = urlparse(repo)
    repo = p.path.strip("/")

    api_url = f"https://api.github.com/repos/{repo}/pulls"
    page = 1
    carry_on = True
    while carry_on:
        r = httpx.get(
            api_url,
            headers={
                "accept": "application/vnd.github+json",
                "X-GitHub-Api-Version": "2022-11-28",
                "Authorization": f"Bearer {gh_token}",
            },
            params={
                "base": base,
                "direction": "desc",
                "state": "all",
                "sort": "created",
                "page": page,
            },
        )

        if r.status_code > 399:
            logger.debug(f"failed to get PR for repo '{repo}': {r.json()}")
            raise ActivityFailed(f"failed to retrieve PR for repo '{repo}'")

        pulls = r.json()
        if not pulls:
            break

        page = page + 1
        for pull in pulls:
            closed_at = pull["closed_at"]
            if closed_at:
                closed_dt = datetime.strptime(closed_at, "%Y-%m-%dT%H:%M:%SZ")
                if closed_dt < start_period:
                    break
                total_closed_during_period += 1

            created_at = pull["created_at"]
            if created_at:
                created_dt = datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%SZ")
                if created_dt >= start_period:
                    total_opened_during_period += 1
                elif only_opened_and_closed_during_window:
                    carry_on = False

                if not closed_at:
                    total_opened += 1

    total = total_opened
    if only_opened_and_closed_during_window:
        total = total_opened_during_period

    if total == 0 and total_closed_during_period > 0:
        ratio = 100.0
    elif total == 0:
        ratio = 0.0
    else:
        ratio = (total_closed_during_period * 100.0) / total

    logger.debug(
        f"Found {total} PRs still opened, "
        f"{total_opened_during_period} opened during the window and "
        f"{total_closed_during_period} closed during the window. "
        f"Leading to a ratio: {ratio}%"
    )

    return ratio


def pr_duration(
    repo: str,
    base: str = "main",
    window: Optional[str] = "5d",
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> List[float]:
    """
    Get a list of opened pull-requests durations.

    If you don't set a window (by setting `window` to `None`), then it returns
    the duration of all PRs that were ever opened in this repository. Otherwise,
    only return the durations for PRs that were opened or closed within that
    window.

    The `repo` should be given as `owner/repo` and the window should be given
    as a pattern like this: `<int>s|m|d|w` (seconds, minutes, days, weeks).
    """
    secrets = secrets or {}
    gh_token = secrets.get("github", {}).get("token")
    gh_token = os.getenv("GITHUB_TOKEN", gh_token)

    if not gh_token:
        raise InvalidActivity(
            "`pr_opened_duration` requires a github token as a secret or via "
            "the GITHUB_TOKEN environment variable"
        )

    today = datetime.today()
    start_period = None
    if window:
        duration = parse_duration(window)
        start_period = today - duration

    durations = []
    p = urlparse(repo)
    repo = p.path.strip("/")

    api_url = f"https://api.github.com/repos/{repo}/pulls"
    page = 1
    carry_on = True
    while carry_on:
        r = httpx.get(
            api_url,
            headers={
                "accept": "application/vnd.github+json",
                "X-GitHub-Api-Version": "2022-11-28",
                "Authorization": f"Bearer {gh_token}",
            },
            params={
                "base": base,
                "direction": "desc",
                "state": "all",
                "sort": "created",
                "page": page,
            },
        )

        if r.status_code > 399:
            logger.debug(f"failed to get PR for repo '{repo}': {r.json()}")
            raise ActivityFailed(f"failed to retrieve PR for repo '{repo}'")

        pulls = r.json()
        if not pulls:
            break

        closed_at = created_at = None
        page = page + 1
        for pull in pulls:
            closed_at = pull["closed_at"]
            if closed_at:
                closed_dt = datetime.strptime(closed_at, "%Y-%m-%dT%H:%M:%SZ")
                if start_period and closed_dt < start_period:
                    continue

            created_at = pull["created_at"]
            if created_at:
                created_dt = datetime.strptime(created_at, "%Y-%m-%dT%H:%M:%SZ")
                if start_period and created_dt < start_period:
                    continue
            else:
                continue

            # deal with PRs that aren't closed yet
            if not closed_at:
                closed_dt = today

            durations.append((closed_dt - created_dt).total_seconds())

    return durations



import os
from typing import Any

import httpx
from chaoslib.types import Configuration, Experiment, Journal, Secrets
from logzero import logger

from chaosreliably.controls import find_extension_by_name

__all__ = ["after_experiment_control"]
OPENAI_URL = "https://api.openai.com/v1/chat/completions"


def after_experiment_control(
    context: Experiment,
    state: Journal,
    openai_model: str = "gpt-3.5-turbo",
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs: Any,
) -> None:
    extension = find_extension_by_name(context, "chatgpt")
    if not extension:
        return None

    secrets = secrets or {}
    openapi_secrets = secrets.get("openai", {})

    org = openapi_secrets.get("org") or os.getenv("OPENAI_ORG")
    if not org:
        logger.warning("Cannot call OpenAPI: missing org")
        return None

    key = openapi_secrets.get("key") or os.getenv("OPENAI_API_KEY")
    if not key:
        logger.warning("Cannot call OpenAPI: missing secret key")
        return None

    logger.debug(
        f"Asking OpenAPI for chat completions using model '{openai_model}'"
    )

    messages = extension.get("messages", [])
    try:
        r = httpx.post(
            OPENAI_URL,
            timeout=60,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {key}",
                "OpenAI-Organization": org,
            },
            json={
                "model": openai_model,
                "temperature": 0.2,
                "messages": messages,
            },
        )
    except httpx.ReadTimeout:
        logger.debug("OpenAI took too long to respond unfortunately")
    else:
        if r.status_code == 200:
            extension["results"] = r.json()



from typing import Any, Dict, List, Optional

__all__ = ["has_error_budget_left"]


def has_error_budget_left(
    name: str, value: Optional[List[Dict[str, Any]]] = None
) -> bool:
    if not value:
        return False

    for item in value:
        if item.get("name") == name:
            return float(item["error_budget_burn_rate"]) < 1.0

    return False



import json
import os.path
from typing import Optional, cast

__all__ = [
    "load_test_result_field_should_be",
    "load_test_result_field_should_be_less_than",
    "load_test_result_field_should_be_greater_than",
]


def load_test_result_field_should_be(
    result_filepath: str,
    field: str,
    expect: int,
    result_item_name: Optional[str] = None,
    pass_if_file_is_missing: bool = True,
) -> bool:
    """
    Reads a load test result and compares the field's value to the expected
    given value.

    If the load test runs against many endpoint, specify which one must be
    validated by setting the `result_item_name` to match the `name` field.
    """
    if not os.path.exists(result_filepath):
        return pass_if_file_is_missing

    with open(result_filepath) as f:
        data = json.load(f)

    if result_item_name:
        for item in data:
            if result_item_name == item["name"]:
                return cast(bool, item[field] == expect)
    else:
        return cast(bool, data[0][field] == expect)

    return False


def load_test_result_field_should_be_less_than(
    result_filepath: str,
    field: str,
    expect: int,
    result_item_name: Optional[str] = None,
    pass_if_file_is_missing: bool = True,
) -> bool:
    """
    Reads a load test result and compares the field's value to less than the
    expected given value.

    If the load test runs against many endpoint, specify which one must be
    validated by setting the `result_item_name` to match the `name` field.
    """
    if not os.path.exists(result_filepath):
        return pass_if_file_is_missing

    with open(result_filepath) as f:
        data = json.load(f)

    if result_item_name:
        for item in data:
            if result_item_name == item["name"]:
                return cast(bool, item[field] < expect)
    else:
        return cast(bool, data[0][field] < expect)

    return False


def load_test_result_field_should_be_greater_than(
    result_filepath: str,
    field: str,
    expect: int,
    result_item_name: Optional[str] = None,
    pass_if_file_is_missing: bool = True,
) -> bool:
    """
    Reads a load test result and compares the field's value to greater than the
    expected given value.

    If the load test runs against many endpoint, specify which one must be
    validated by setting the `result_item_name` to match the `name` field.
    """
    if not os.path.exists(result_filepath):
        return pass_if_file_is_missing

    with open(result_filepath) as f:
        data = json.load(f)

    if result_item_name:
        for item in data:
            if result_item_name == item["name"]:
                return cast(bool, item[field] > expect)
    else:
        return cast(bool, data[0][field] > expect)

    return False



import httpx

__all__ = ["measure_response_time"]


def measure_response_time(url: str) -> float:
    """
    Measure the response time of the GET request to the given URL.
    """
    with httpx.Client() as c:
        r = c.get(url)
        return r.elapsed.total_seconds()



from chaosreliably import __version__, discover


def test_that_discover_returns_correct_discovery() -> None:
    discovery = discover()
    assert discovery["extension"]["name"] == "chaostoolkit-reliably"
    assert discovery["extension"]["version"] == __version__
    names = [activity["name"] for activity in discovery["activities"]]
    assert len(names) == 22



import io
import json
import os
import secrets
import threading
import time
from copy import deepcopy
from datetime import datetime, timezone
from logging import Formatter, StreamHandler
from typing import Any, Dict, Optional, cast

from chaoslib.exit import exit_gracefully, exit_ungracefully
from chaoslib.run import EventHandlerRegistry, RunEventHandler
from chaoslib.types import (
    Activity,
    Configuration,
    Experiment,
    Journal,
    Run,
    Schedule,
    Secrets,
    Settings,
)
from logzero import logger

from chaosreliably import RELIABLY_HOST, get_session
from chaosreliably.activities.pauses import reset as reset_activity_pause

__all__ = ["configure_control"]


class ReliablyHandler(RunEventHandler):  # type: ignore
    def __init__(
        self,
        org_id: str,
        exp_id: str,
        experiment: Experiment,
    ) -> None:
        RunEventHandler.__init__(self)
        self.org_id = org_id
        self.exp_id = exp_id
        self.exec_id = None  # type: Optional[str]
        self.started = datetime.utcnow().replace(tzinfo=timezone.utc)
        self.experiment = deepcopy(experiment)
        self.journal = None  # type: Journal
        self.current_activities = []  # type: ignore

        self.should_stop = threading.Event()
        self.should_pause = threading.Event()
        self.check_lock = threading.Lock()
        self.pause_duration = 0
        self.paused_by_user = ""
        self.paused_by_user_id = ""
        self.paused = False
        self.check_for_user_state = None  # type: Optional[threading.Thread]

        self.stream = io.StringIO()
        self.log_handler = StreamHandler(stream=self.stream)
        self.log_handler.setFormatter(
            Formatter("[%(asctime)s] - %(levelname)s - %(message)s")
        )

        logger.addHandler(self.log_handler)

    def _check(
        self,
        org_id: str,
        exp_id: str,
        exec_id: str,
        configuration: Configuration,
        secrets: Secrets,
    ) -> None:
        logger.debug("Starting Reliably state checker now")

        url = f"/{org_id}/experiments/{exp_id}/executions/{exec_id}/state"
        while not self.should_stop.is_set():
            with get_session(configuration, secrets) as session:
                r = session.get(url)
                if r.status_code > 399:
                    logger.debug(f"Failed to retrieve state: {r.json()}")
                    continue

            state = r.json()
            if state:
                if state["current"] == "terminate":
                    user = state.get("user", {})
                    username = user.get("name", "")
                    self.paused_by_user = ""
                    self.paused_by_user_id = ""

                    logger.info(
                        "Execution state was changed to `terminate` "
                        f"by {username}. "
                        "Exiting now..."
                    )

                    reset_activity_pause()

                    self.extension["termination"] = {
                        "timestamp": datetime.utcnow()
                        .replace(tzinfo=timezone.utc)
                        .isoformat(),
                        "skip_rollbacks": state["skip_rollbacks"],
                    }

                    if state["skip_rollbacks"]:
                        exit_ungracefully()
                    else:
                        exit_gracefully()
                elif state["current"] == "pause":
                    duration = state["duration"]
                    with self.check_lock:
                        if not self.paused:
                            user = state.get("user", {})
                            self.paused_by_user = user.get("name", "")
                            self.paused_by_user_id = user.get("id", "")

                            m = (
                                "Execution state was changed to `pause`. "
                                "Pausing the execution of the current "
                                "activity"
                            )
                            if self.paused_by_user:
                                m = f"{m}, requested by {self.paused_by_user},"

                            if duration:
                                m = f"{m} for {duration}s"
                            else:
                                m = f"{m} indefinitely"
                            m = f"{m} or until state changes back to `resume`"
                            logger.info(m)
                            self.paused = True
                            self.pause_duration = state["duration"]
                            self.should_pause.set()
                elif state["current"] == "resume":
                    reset_activity_pause()

                    with self.check_lock:
                        if self.paused:
                            user = state.get("user", {})
                            username = user.get("name", "")
                            self.paused_by_user = ""
                            self.paused_by_user_id = ""

                            logger.info(
                                "Execution state was changed to `resume` "
                                f"by {username}"
                            )

            now = time.time()
            later = now + 10
            while time.time() < later:
                time.sleep(0.5)
                if self.should_stop.is_set():
                    logger.debug("Stopping Reliably state checker now")
                    break

    def running(
        self,
        experiment: Experiment,
        journal: Journal,
        configuration: Configuration,
        secrets: Secrets,
        schedule: Schedule,
        settings: Settings,
    ) -> None:
        self.experiment = experiment
        self.configuration = configuration
        self.secrets = secrets
        self.journal = journal

        try:
            result = create_run(
                self.org_id,
                self.exp_id,
                experiment,
                journal,
                self.configuration,
                self.secrets,
            )

            if result:
                payload = result
                self.extension = get_reliably_extension_from_journal(journal)

                self.exec_id = payload["id"]
                logger.info(f"Reliably execution: {self.exec_id}")

                host = self.secrets.get(
                    "host", os.getenv("RELIABLY_HOST", RELIABLY_HOST)
                )

                url = f"https://{host}/executions/view/?id={self.exec_id}&exp={self.exp_id}"  # noqa
                self.extension["execution_url"] = url

                add_runtime_extra(self.extension)
                set_plan_status(
                    self.org_id,
                    "running",
                    None,
                    self.configuration,
                    self.secrets,
                )

                set_execution_state(
                    self.org_id,
                    self.exp_id,
                    self.exec_id,
                    {
                        "current": "running",
                        "started_on": self.started.isoformat(),
                    },
                    self.configuration,
                    self.secrets,
                )

                self.check_for_user_state = threading.Thread(
                    None,
                    self._check,
                    args=(
                        self.org_id,
                        self.exp_id,
                        self.exec_id,
                        configuration,
                        secrets,
                    ),
                )
                self.check_for_user_state.start()
        except Exception as ex:
            set_plan_status(
                self.org_id, "error", str(ex), self.configuration, self.secrets
            )

    def finish(self, journal: Journal) -> None:
        logger.info("Finishing Reliably execution...")
        self.should_stop.set()
        if self.check_for_user_state:
            self.check_for_user_state.join(timeout=3)
            self.check_for_user_state = None

        logger.removeHandler(self.log_handler)
        self.log_handler.flush()

        log = self.stream.getvalue()
        self.stream.close()

        self.current_activities = []

        try:
            complete_run(
                self.org_id,
                self.exp_id,
                self.exec_id,
                journal,
                log,
                self.configuration,
                self.secrets,
            )
            set_plan_status(
                self.org_id,
                "completed",
                None,
                self.configuration,
                self.secrets,
            )
        except Exception as ex:
            set_plan_status(
                self.org_id, "error", str(ex), self.configuration, self.secrets
            )
        finally:
            set_execution_state(
                self.org_id,
                self.exp_id,
                self.exec_id,
                {
                    "current": "finished",
                    "status": journal.get("status"),
                    "deviated": journal.get("deviated"),
                },
                self.configuration,
                self.secrets,
            )

            self.experiment = (
                self.configuration
            ) = self.secrets = self.journal = None

        logger.info("Finished Reliably execution. Bye!")

    def start_hypothesis_before(self, experiment: Experiment) -> None:
        with self.check_lock:
            self.current_activities = experiment.get(
                "steady-state-hypothesis", {}
            ).get("probes", [])

    def start_hypothesis_after(self, experiment: Experiment) -> None:
        ssh = experiment.get("steady-state-hypothesis")
        if ssh:
            with self.check_lock:
                self.current_activities = ssh.get("probes", [])
                experiment["steady-state-hypothesis"][
                    "probes"
                ] = self.current_activities

    def start_method(self, experiment: Experiment) -> None:
        with self.check_lock:
            self.current_activities = experiment.get("method", [])

    def start_rollbacks(self, experiment: Experiment) -> None:
        with self.check_lock:
            self.current_activities = experiment.get("rollbacks", [])

    def start_activity(self, activity: Activity) -> None:
        name = activity.get("name")

        with self.check_lock:
            if self.paused:
                pause = {
                    "before_activity": name,
                    "start": datetime.utcnow()
                    .replace(tzinfo=timezone.utc)
                    .isoformat(),
                    "duration": self.pause_duration,
                }
                self.extension["pauses"].append(pause)

        send_journal(
            self.org_id,
            self.exp_id,
            self.exec_id,
            self.journal,
            self.configuration,
            self.secrets,
        )

    def activity_completed(self, activity: Activity, run: Run) -> None:
        name = activity.get("name")

        if self.should_pause.is_set():
            self.should_pause.clear()

            logger.info(f"Adding a pause activity after '{name}'")

            duration = 0
            with self.check_lock:
                duration = self.pause_duration

            index = self.current_activities.index(activity)
            with self.check_lock:
                self.current_activities.insert(
                    index + 1,
                    make_user_pause(
                        duration, self.paused_by_user, self.paused_by_user_id
                    ),
                )
        else:
            with self.check_lock:
                if self.paused:
                    self.paused = False

                    self.extension["pauses"][-1]["end"] = (
                        datetime.utcnow()
                        .replace(tzinfo=timezone.utc)
                        .isoformat()
                    )

                    set_execution_state(
                        self.org_id,
                        self.exp_id,
                        self.exec_id,
                        {
                            "current": "running",
                            "started_on": self.started.isoformat(),
                        },
                        self.configuration,
                        self.secrets,
                    )

                    logger.info("No longer paused")

        send_journal(
            self.org_id,
            self.exp_id,
            self.exec_id,
            self.journal,
            self.configuration,
            self.secrets,
        )


def configure_control(
    experiment: Experiment,
    event_registry: EventHandlerRegistry,
    exp_id: str,
    org_id: str,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs: Any,
) -> None:
    logger.debug("Configure Reliably's experiment control")
    event_registry.register(ReliablyHandler(org_id, exp_id, experiment))


###############################################################################
# Private functions
###############################################################################
def make_user_pause(
    pause_duration: int = 0, username: str = "", user_id: str = ""
) -> Activity:
    return {
        "type": "action",
        "name": f"reliably-pause-{secrets.token_hex(4)}",
        "provider": {
            "type": "python",
            "module": "chaosreliably.activities.pauses",
            "func": "pause_execution",
            "arguments": {
                "duration": pause_duration,
                "username": username,
                "user_id": user_id,
            },
        },
    }


def create_run(
    org_id: str,
    exp_id: str,
    experiment: Experiment,
    state: Journal,
    configuration: Configuration,
    secrets: Secrets,
) -> Optional[Dict[str, Any]]:
    with get_session(configuration, secrets) as session:
        resp = session.post(
            f"/{org_id}/experiments/{exp_id}/executions",
            json={"result": json.dumps(state)},
        )
        if resp.status_code == 201:
            return cast(Dict[str, Any], resp.json())
    return None


def complete_run(
    org_id: str,
    exp_id: str,
    execution_id: Optional[str],
    state: Journal,
    log: str,
    configuration: Configuration,
    secrets: Secrets,
) -> Optional[Dict[str, Any]]:
    with get_session(configuration, secrets) as session:
        resp = session.put(
            f"/{org_id}/experiments/{exp_id}/executions/{execution_id}/results",
            json={"result": json.dumps(state), "log": log},
        )
        if resp.status_code != 200:
            logger.error("Failed to update results on server")
    return None


def send_journal(
    org_id: str,
    exp_id: str,
    execution_id: Optional[str],
    state: Journal,
    configuration: Configuration,
    secrets: Secrets,
) -> Optional[Dict[str, Any]]:
    with get_session(configuration, secrets) as session:
        resp = session.put(
            f"/{org_id}/experiments/{exp_id}/executions/{execution_id}/results",
            json={"result": json.dumps(state)},
        )
        if resp.status_code != 200:
            logger.error("Failed to update results on server")
    return None


def get_reliably_extension_from_journal(journal: Journal) -> Dict[str, Any]:
    experiment = journal.get("experiment")
    extensions = experiment.setdefault("extensions", [])
    for extension in extensions:
        if extension["name"] == "reliably":
            return cast(Dict[str, Any], extension)

    extension = {"name": "reliably", "pauses": [], "termination": None}
    extensions.append(extension)
    return cast(Dict[str, Any], extension)


def add_runtime_extra(extension: Dict[str, Any]) -> None:
    extra = os.getenv("RELIABLY_EXECUTION_EXTRA", "[]")

    try:
        extension["extra"] = json.loads(extra)
    except Exception:
        logger.debug("Failed to parse RELIABLY_EXECUTION_EXTRA")
        extension["extra"] = []

    if "CI" in os.environ:
        repo = os.getenv("GITHUB_REPOSITORY")
        gh = os.getenv("GITHUB_SERVER_URL")
        run_id = os.getenv("GITHUB_RUN_ID")

        if repo and gh:
            extension["extra"].append(
                {
                    "type": "url",
                    "provider": "github",
                    "topic": "repo",
                    "value": f"{gh}/{repo}",
                }
            )

        if repo and gh and run_id:
            extension["extra"].append(
                {
                    "type": "url",
                    "provider": "github",
                    "topic": "run",
                    "value": f"{gh}/{repo}/actions/runs/{run_id}",
                }
            )


def set_plan_status(
    org_id: str,
    status: str,
    message: Optional[str],
    configuration: Configuration,
    secrets: Secrets,
) -> None:
    plan_id = os.getenv("RELIABLY_PLAN_ID")
    if not plan_id:
        return None

    logger.debug(f"Sending plan '{plan_id}' status: {status}")

    with get_session(configuration, secrets) as session:
        r = session.put(
            f"/{org_id}/plans/{plan_id}/status",
            json={"status": status, "error": message},
        )
        if r.status_code > 399:
            logger.debug(
                f"Failed to set plan status: {r.status_code}: {r.json()}"
            )


def set_execution_state(
    org_id: str,
    exp_id: str,
    exec_id: Optional[str],
    state: Dict[str, Any],
    configuration: Configuration,
    secrets: Secrets,
) -> None:
    if not exec_id:
        return None

    state["plan_id"] = os.getenv("RELIABLY_PLAN_ID")

    with get_session(configuration, secrets) as session:
        r = session.put(
            f"/{org_id}/experiments/{exp_id}/executions/{exec_id}/state",
            json=state,
        )
        if r.status_code > 399:
            logger.debug(
                f"Failed to set execution state: {r.status_code}: {r.json()}"
            )


def to_datetime(ts: str) -> datetime:
    return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%f").replace(
        tzinfo=timezone.utc
    )



from typing import List, Optional, Union, cast

from boltons import statsutils
from chaoslib.exceptions import InvalidActivity
from logzero import logger

from chaosreliably import parse_duration

__all__ = [
    "ratio_under",
    "ratio_above",
    "percentile_under",
    "ratio_under_or_equal",
    "ratio_above_or_equal",
]


def ratio_under(target: float, value: float = 0.0) -> bool:
    """
    Validates the ratio returned by a probe is strictly below the `target`.
    """
    logger.debug(f"Verify that ratio is below: {target}")
    return value < target


def ratio_above(target: float, value: float = 0.0) -> bool:
    """
    Validates the ratio returned by a probe is strictly greater than the
    `target`.
    """
    logger.debug(f"Verify that ratio is above: {target}")
    return value > target


def ratio_under_or_equal(target: float, value: float = 0.0) -> bool:
    """
    Validates the ratio returned by a probe is below the `target`.
    """
    logger.debug(f"Verify that ratio is below: {target}")
    return value <= target


def ratio_above_or_equal(target: float, value: float = 0.0) -> bool:
    """
    Validates the ratio returned by a probe is greater than the
    `target`.
    """
    logger.debug(f"Verify that ratio is above: {target}")
    return value >= target


def percentile_under(
    percentile: float,
    duration: str = "1d",
    value: Optional[List[Union[int, float]]] = None,
) -> bool:
    """
    Computes that the values under `percentile` are below the given duration.

    For instance, for PR durations, this could be helpful to understand that
    99% of them were closed in less than the given duration.

    ```python
    v = pr_duration("chaostoolkit/chaostoolkit", "master", window=None)
    p = percentile_under(0.99, duration="1d", value=v)
    ```
    """
    if not (0.0 <= percentile <= 1.0):
        raise InvalidActivity(
            "`percentile` of the `percentile_under` tolerance "
            "must be below 0 and 99"
        )

    if not value:
        return True

    d = parse_duration(duration).total_seconds()
    s = statsutils.Stats(value)
    q = cast(float, s.get_quantile(percentile))

    logger.debug(f"Stats summary:\n{s.describe(format='text')}")

    return q <= d



from logzero import logger

__all__ = ["response_time_must_be_under"]


def response_time_must_be_under(latency: float, value: float = 0.0) -> bool:
    """
    Validates the response time is under the given latency.

    Use this as the tolerance of the
    `chaosreliably.activities.http.probes.measure_response_time` probe.
    """
    logger.debug(f"Verify that response time is under: {latency}")
    return value <= latency



from chaosreliably import get_auth_info


def test_using_secret_only() -> None:
    auth_info = get_auth_info(None, {"token": "78890", "host": "reliably.dev"})
    assert auth_info["token"] == "78890"
    assert auth_info["host"] == "reliably.dev"


def test_missing_host_from_secrets() -> None:
    auth_info = get_auth_info(
        {},
        {"token": "78890", "org": "an-org"},
    )
    assert auth_info["token"] == "78890"
    assert auth_info["host"] == "app.reliably.com"



import json
import os
import os.path
import pkgutil
import shutil
import subprocess  # nosec
import tempfile
from typing import Any, Dict, Optional, cast
from urllib.parse import urlparse

from chaoslib import decode_bytes
from chaoslib.exceptions import ActivityFailed, InvalidActivity
from chaoslib.types import Configuration, Secrets
from logzero import logger

__all__ = ["inject_gradual_traffic_into_endpoint"]


def inject_gradual_traffic_into_endpoint(
    endpoint: str,
    step_duration: int = 5,
    step_additional_vu: int = 1,
    vu_per_second_rate: int = 1,
    test_duration: int = 30,
    results_json_filepath: Optional[str] = None,
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> Dict[str, Any]:
    """
    Load traffic into the given `endpoint`. Uses an approach that creates an
    incremental load into the endpoint rather than swarming it. The point of
    this action is to ensure your endpoint is active while you perform another
    action. This you means you likely want to run this action in the
    `background`.

    You may set a bearer token if your application uses one to authenticate.
    Pass `test_bearer_token` as a secret key in the `secrets` payload.

    This action return a dictionary payload of the load test results.
    """
    u = urlparse(endpoint)
    if not u.scheme or not u.netloc:
        raise InvalidActivity("endpoint must be a proper url")

    script = pkgutil.get_data(
        "chaosreliably", "activities/load/scripts/step_load_test.py"
    )
    if not script:
        raise ActivityFailed("failed to locate load-test script")

    locust_path = shutil.which("locust")
    if not locust_path:
        raise ActivityFailed("missing load test dependency")

    env = {
        "RELIABLY_LOCUST_ENDPOINT": endpoint,
        "RELIABLY_LOCUST_STEP_TIME": str(step_duration),
        "RELIABLY_LOCUST_STEP_LOAD": str(step_additional_vu),
        "RELIABLY_LOCUST_SPAWN_RATE": str(vu_per_second_rate),
        "RELIABLY_LOCUST_TIME_LIMIT": str(test_duration),
    }

    secrets = secrets or {}
    test_bearer_token = secrets.get("test_bearer_token")
    if test_bearer_token:
        env["RELIABLY_LOCUST_ENDPOINT_TOKEN"] = test_bearer_token

    results = {}

    with tempfile.TemporaryDirectory() as d:
        locustfile_path = os.path.join(d, "locustfile.py")
        with open(locustfile_path, mode="wb") as f:
            f.write(script)

        cmd = [
            locust_path,
            "--host",
            "localhost:8089",
            "--locustfile",
            locustfile_path,
            "--json",
            "--headless",
            "--loglevel",
            "INFO",
            "--exit-code-on-error",
            "0",
        ]
        try:
            p = subprocess.run(  # nosec
                cmd,
                timeout=test_duration + 5,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=env,
                shell=False,
                cwd=d,
            )
            os.remove(locustfile_path)
        except KeyboardInterrupt:
            logger.debug(
                "Caught SIGINT signal while running locust. Ignoring it."
            )
        except subprocess.TimeoutExpired:
            raise ActivityFailed("load test took too long to complete")
        else:
            stdout = decode_bytes(p.stdout)
            stderr = decode_bytes(p.stderr)

            logger.debug(f"locust exit code: {p.returncode}")
            logger.debug(f"locust stderr: {stderr}")

            if results_json_filepath:
                with open(results_json_filepath, "w") as f:
                    f.write(stdout)

            try:
                results = json.loads(stdout)
            except json.decoder.JSONDecodeError:
                logger.error("failed to parse locust results")

    return cast(Dict[str, Any], results)



import threading

from logzero import logger

__all__ = ["pause_execution"]

CURRENT_PAUSE = threading.Event()


def pause_execution(
    duration: int = 0,
    username: str = "",
    user_id: str = "",
) -> None:
    """
    Pause the execution of the experiment until the resume state has been
    received.
    """
    m = "Pausing activity"
    if duration:
        m = f"{m} for {duration}s or"
    m = f"{m} until the execution is resumed"
    logger.info(m)

    if CURRENT_PAUSE.wait(duration or None) is False:
        logger.info("Resuming execution...")


###############################################################################
# Private functions
###############################################################################
def reset() -> None:
    global CURRENT_PAUSE
    CURRENT_PAUSE.set()
    CURRENT_PAUSE.clear()



import sys

if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError, version

elif sys.version_info < (3, 8):
    from importlib_metadata import PackageNotFoundError, version


try:
    __version__ = version("chaostoolkit-reliably")
except PackageNotFoundError:
    __version__ = "unknown"



from typing import Any, Optional

from chaoslib.run import EventHandlerRegistry
from chaoslib.types import Configuration, Experiment, Secrets

from . import initialize, register, run_all

__all__ = ["configure_control"]


def configure_control(
    event_registry: EventHandlerRegistry,
    url: str,
    frequency: float,
    auth: Optional[str] = None,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs: Any,
) -> None:
    initialize(event_registry)
    register(url=url, auth=auth, frequency=frequency)


def before_experiment_control(
    context: Experiment,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs: Any,
) -> None:
    run_all(context, configuration, secrets)



import os
import secrets
from typing import Any, Dict, Union

from chaoslib.types import Activity, Configuration, Experiment, Secrets
from logzero import logger

from chaosreliably.types import AutoPause

__all__ = ["configure_control"]


def configure_control(
    experiment: Experiment,
    autopause: AutoPause,
    configuration: Configuration = None,
    secrets: Secrets = None,
    **kwargs: Any,
) -> None:
    logger.debug("Configure Reliably's autopause control")

    amend_experiment_for_autopauses(experiment, autopause)


###############################################################################
# Private functions
###############################################################################
def amend_experiment_for_autopauses(
    experiment: Experiment, autopause: AutoPause
) -> None:
    method = experiment.get("method")
    if method and "method" in autopause:
        p = autopause["method"]
        if is_enabled(p.get("actions", {}).get("enabled")):
            pause_duration = get_duration(
                p.get("actions", {}).get("pause_duration", 0)
            )

            activities = method[:]
            for index, activity in enumerate(activities):
                index = method.index(activity)
                if activity["type"] == "action":
                    method.insert(index + 1, make_pause(pause_duration))

        if is_enabled(p.get("probes", {}).get("enabled")):
            pause_duration = get_duration(
                p.get("probes", {}).get("pause_duration", 0)
            )

            activities = method[:]
            for index, activity in enumerate(activities):
                index = method.index(activity)
                if activity["type"] == "probe":
                    method.insert(index + 1, make_pause(pause_duration))

    ssh_probes = experiment.get("steady-state-hypothesis", {}).get("probes")
    if ssh_probes and "steady-state-hypothesis" in autopause:
        p = autopause["steady-state-hypothesis"]
        if is_enabled(p["enabled"]):
            pause_duration = get_duration(p.get("pause_duration", 0))

            activities = ssh_probes[:]
            for index, activity in enumerate(activities):
                index = ssh_probes.index(activity)
                ssh_probes.insert(index + 1, make_pause(pause_duration))

    rollbacks = experiment.get("rollbacks")
    if rollbacks and "rollbacks" in autopause:
        p = autopause["rollbacks"]
        if is_enabled(p["enabled"]):
            pause_duration = get_duration(p.get("pause_duration", 0))

            activities = rollbacks[:]
            for index, activity in enumerate(activities):
                index = rollbacks.index(activity)
                rollbacks.insert(index + 1, make_pause(pause_duration))


def make_pause(pause_duration: float = 0) -> Activity:
    return {
        "type": "action",
        "name": f"reliably-autopause-{secrets.token_hex(4)}",
        "provider": {
            "type": "python",
            "module": "chaosreliably.activities.pauses",
            "func": "pause_execution",
            "arguments": {"duration": pause_duration},
        },
    }


def is_enabled(value: Union[bool, Dict[str, str]]) -> bool:
    if not value:
        return False

    if isinstance(value, bool):
        return value

    if value.get("type") == "env":
        if os.getenv(value["key"], "") not in ("", "f", "false", "0", "False"):
            return True

        return False

    return False


def get_duration(value: Union[float, int, Dict[str, str]]) -> float:
    if not value:
        return 0.0

    if isinstance(value, (float, int)):
        return float(value)

    if value.get("type") == "env":
        return float(os.getenv(value["key"]) or "0")

    return 0.0



import time
import httpx
import respx

from chaoslib.run import EventHandlerRegistry
from chaosreliably.activities.safeguard.probes import call_endpoint
from chaosreliably.controls import ReliablyGuardian, ReliablySafeguardHandler, initialize, register, run_all


class TestReliablyGuardian(ReliablyGuardian):
    def _exit(self) -> None:
        time.sleep(2)


def test_prechecks_run_interrupts_execution(respx_mock):
    url = "https://example.com/try-me"

    m = respx_mock.get(url).mock(return_value=httpx.Response(
        200, json={"ok": False, "error": "boom"})
    )
    
    proxy = ReliablySafeguardHandler()
    registry = EventHandlerRegistry()
    initialize(registry, handler=proxy)
    register(url, handler=proxy, guardian_class=TestReliablyGuardian)

    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": []
    }
    journal = {
        "experiment": experiment
    }

    try:
        run_all(experiment, None, None, handler=proxy)
        time.sleep(2)
    finally:
        registry.finish(journal)
        registry.handlers.clear()

    respx_mock.calls.assert_called_once()
    assert proxy.guardians[0].guardian.interrupted is True


def test_safeguard_expects_a_200(respx_mock):
    url = "https://example.com/try-me"

    respx_mock.get(url).mock(return_value=httpx.Response(400))
    assert call_endpoint(url) is False

    respx_mock.get(url).mock(return_value=httpx.Response(200, json={"ok": True}))
    assert call_endpoint(url) is True


def test_safeguard_not_ok_expects_error_message(respx_mock):
    url = "https://example.com/try-me"

    respx_mock.get(url).mock(return_value=httpx.Response(400))
    assert call_endpoint(url) is False

    respx_mock.get(url).mock(return_value=httpx.Response(200, json={
        "ok": False,
        "error": "boom"
    }))
    assert call_endpoint(url) is False


def test_prechecks_run_once(respx_mock):
    url = "https://example.com/try-me"

    m = respx_mock.get(url).mock(return_value=httpx.Response(200, json={"ok": True}))
    
    proxy = ReliablySafeguardHandler()
    registry = EventHandlerRegistry()
    initialize(registry, handler=proxy)
    register(url, handler=proxy, guardian_class=TestReliablyGuardian)

    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": []
    }
    journal = {
        "experiment": experiment
    }

    try:
        run_all(experiment, None, None, handler=proxy)
        time.sleep(2)
    finally:
        registry.finish(journal)
        registry.handlers.clear()

    respx_mock.calls.assert_called_once()
    assert proxy.guardians[0].guardian.interrupted is False


def test_safeguard_run_periodically(respx_mock):
    url = "https://example.com/try-me"

    m = respx_mock.get(url).mock(side_effect=[
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"})
    ])

    proxy = ReliablySafeguardHandler()
    registry = EventHandlerRegistry()
    initialize(registry, handler=proxy)
    register(url, frequency=0.5, handler=proxy, guardian_class=TestReliablyGuardian)

    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": []
    }
    journal = {
        "experiment": experiment
    }

    try:
        run_all(experiment, None, None, handler=proxy)
        time.sleep(2)
    finally:
        registry.finish(journal)
        registry.handlers.clear()

    respx_mock.calls.calls.call_count > 1
    assert proxy.guardians[0].guardian.interrupted is True


def test_safeguard_run_interrupts_execution(respx_mock):
    url = "https://example.com/try-me"

    m = respx_mock.get(url).mock(side_effect=[
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"})
    ])
    
    proxy = ReliablySafeguardHandler()
    registry = EventHandlerRegistry()
    initialize(registry, handler=proxy)
    register(url, frequency=0.5, handler=proxy, guardian_class=TestReliablyGuardian)

    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": []
    }
    journal = {
        "experiment": experiment
    }

    try:
        run_all(experiment, None, None, handler=proxy)
        time.sleep(2)
    finally:
        registry.finish(journal)
        registry.handlers.clear()
        time.sleep(2)

    assert respx_mock.calls.call_count > 1
    assert proxy.guardians[0].guardian.interrupted is True


def test_safeguard_can_be_many(respx_mock):
    url = "https://example.com/try-me"

    m = respx.get(url).mock(side_effect=[
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"}),
        httpx.Response(200, json={"ok": False, "error": "boom"})
    ])
    
    url2 = "https://example.com/try-me-as-well"

    m = respx.get(url2).mock(side_effect=[
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": True}),
        httpx.Response(200, json={"ok": True}),
    ])
    
    proxy = ReliablySafeguardHandler()
    registry = EventHandlerRegistry()
    initialize(registry, handler=proxy)
    register(url, frequency=0.5, handler=proxy, guardian_class=TestReliablyGuardian)
    register(url2, frequency=0.5, handler=proxy, guardian_class=TestReliablyGuardian)

    experiment = {
        "title": "an experiment",
        "description": "n/a",
        "method": []
    }
    journal = {
        "experiment": experiment
    }

    try:
        run_all(experiment, None, None, handler=proxy)
        time.sleep(2)
    finally:
        registry.finish(journal)
        registry.handlers.clear()

    assert respx_mock.calls.call_count > 1
    assert proxy.guardians[0].guardian.interrupted is True

    assert respx_mock.calls.call_count > 1
    assert proxy.guardians[1].guardian.interrupted is False



import socket
import ssl
from typing import Any, Dict

from chaoslib.exceptions import ActivityFailed
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

__all__ = ["get_certificate_info"]


# copied mostly from this awesome piece
# https://github.com/dspruell/tls-probe/blob/main/tls_probe.py
def get_certificate_info(host: str, port: int = 443) -> Dict[str, Any]:
    """
    Extract certificate information from the remote connection.
    """
    context = ssl.create_default_context()
    context.verify_mode = ssl.CERT_REQUIRED
    context.check_hostname = True

    with socket.create_connection((host, port)) as sock:
        conn_info = dict(conn={}, cert={"fingerprints": {}, "extensions": {}})
        with context.wrap_socket(sock, server_hostname=host) as secsock:
            cert = secsock.getpeercert(binary_form=True)
            if not cert:
                raise ActivityFailed(
                    f"endpoint {host}:{port} has no certificate"
                )

            cert_data = x509.load_der_x509_certificate(cert, default_backend())

            conn_info["conn"].update(  # type: ignore
                {
                    "version": secsock.version(),
                    "remote_addr": ":".join(
                        [str(_) for _ in secsock.getpeername()]
                    ),
                }
            )

            sig_hash = cert_data.signature_hash_algorithm.name  # type: ignore
            conn_info["cert"].update(  # type: ignore
                {
                    "issuer": cert_data.issuer.rfc4514_string(),
                    "subject": cert_data.subject.rfc4514_string(),
                    "serial_int": cert_data.serial_number,
                    "version": cert_data.version.name,
                    "signature_hash": sig_hash,
                    "not_valid_before": cert_data.not_valid_before.isoformat(),
                    "not_valid_after": cert_data.not_valid_after.isoformat(),
                }
            )

            conn_info["cert"]["fingerprints"].update(  # type: ignore
                {
                    "md5": cert_data.fingerprint(hashes.MD5()).hex(),  # nosec
                    "sha1": cert_data.fingerprint(hashes.SHA1()).hex(),  # nosec
                    "sha256": cert_data.fingerprint(hashes.SHA256()).hex(),
                }
            )

            for ext in cert_data.extensions:
                if ext.oid._name in ("subjectAltName",):
                    names = []  # type: ignore
                    conn_info["cert"]["extensions"].update(  # type: ignore
                        {
                            ext.oid._name: names,
                        }
                    )
                    for g in ext.value._general_names:
                        names.append(g.value)

        return conn_info



import os
from contextlib import contextmanager
from datetime import timedelta
from typing import Dict, Generator, List

import httpx
from chaoslib.discovery.discover import (
    discover_actions,
    discover_activities,
    discover_probes,
    initialize_discovery_result,
)
from chaoslib.types import (
    Configuration,
    DiscoveredActivities,
    Discovery,
    Secrets,
)
from logzero import logger

from .__version__ import __version__

__all__ = ["get_session", "discover", "parse_duration"]
RELIABLY_HOST = "app.reliably.com"


@contextmanager
def get_session(
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> Generator[httpx.Client, None, None]:
    c = configuration or {}
    verify_tls = c.get("reliably_verify_tls", True)
    with_http2 = True if verify_tls else False
    use_http = c.get("reliably_use_http", False)
    scheme = "http" if use_http else "https"
    auth_info = get_auth_info(configuration, secrets)
    headers = {
        "Content-Type": "application/json",
        "Authorization": "Bearer {}".format(auth_info["token"]),
    }

    with httpx.Client(
        verify=verify_tls, http2=with_http2, timeout=30
    ) as client:
        client.headers = httpx.Headers(headers)
        client.base_url = httpx.URL(
            f"{scheme}://{auth_info['host']}/api/v1/organization"
        )
        yield client


def discover(discover_system: bool = True) -> Discovery:
    """
    Discover Reliably capabilities from this extension.
    """
    logger.info("Discovering capabilities from chaostoolkit-reliably")

    discovery = initialize_discovery_result(
        "chaostoolkit-reliably", __version__, "reliably"
    )
    discovery["activities"].extend(load_exported_activities())

    return discovery


###############################################################################
# Private functions
###############################################################################
def get_auth_info(
    configuration: Configuration = None, secrets: Secrets = None
) -> Dict[str, str]:
    reliably_host = None
    reliably_token = None

    reliably_host = secrets.get(
        "host", os.getenv("RELIABLY_HOST", RELIABLY_HOST)
    )

    reliably_token = secrets.get(
        "token", os.getenv("RELIABLY_TOKEN", reliably_token)
    )

    return {"host": reliably_host, "token": reliably_token}


def load_exported_activities() -> List[DiscoveredActivities]:
    """
    Extract metadata from actions, probes and tolerances
    exposed by this extension.
    """
    activities = []
    activities.extend(
        discover_activities(
            "chaosreliably.activities.http.tolerances", "tolerance"
        )
    )
    activities.extend(discover_probes("chaosreliably.activities.http.probes"))
    activities.extend(
        discover_activities(
            "chaosreliably.activities.tls.tolerances", "tolerance"
        )
    )
    activities.extend(discover_probes("chaosreliably.activities.tls.probes"))
    activities.extend(
        discover_activities(
            "chaosreliably.activities.gh.tolerances", "tolerance"
        )
    )
    activities.extend(discover_probes("chaosreliably.activities.gh.probes"))
    activities.extend(discover_probes("chaosreliably.activities.load.probes"))
    activities.extend(discover_actions("chaosreliably.activities.load.actions"))
    activities.extend(discover_probes("chaosreliably.activities.slo.probes"))
    activities.extend(
        discover_activities(
            "chaosreliably.activities.slo.tolerances", "tolerance"
        )
    )
    activities.extend(
        discover_probes("chaosreliably.activities.safeguard.probes")
    )

    return activities


def parse_duration(duration: str) -> timedelta:
    value = int(duration[:-1])
    unit = duration[-1]

    if unit == "s":
        return timedelta(seconds=value)
    elif unit == "m":
        return timedelta(minutes=value)
    elif unit == "d":
        return timedelta(days=value)
    elif unit == "w":
        return timedelta(weeks=value)

    return timedelta(weeks=1)



import logging
import warnings

from logzero import logger


# from slo-egenerator
def reroute_slo_generator_logging() -> None:
    slo_gen_logger = logging.getLogger("slo_generator")
    slo_gen_logger.setLevel(logging.DEBUG)
    slo_gen_logger.propagate = False

    gcp_logger = logging.getLogger("googleapiclient")
    gcp_logger.setLevel(logging.ERROR)
    gcp_logger.propagate = False

    # not ideal here but it's the only way to route these messages to
    # chaostoolkit logger
    for h in logger.handlers:
        slo_gen_logger.addHandler(h)
        gcp_logger.addHandler(h)

    # Ignore Cloud SDK warning when using a user instead of service account
    try:
        # pylint: disable=import-outside-toplevel
        from google.auth._default import (  # type: ignore
            _CLOUD_SDK_CREDENTIALS_WARNING,
        )

        warnings.filterwarnings(
            "ignore", message=_CLOUD_SDK_CREDENTIALS_WARNING
        )
    except ImportError:
        pass


reroute_slo_generator_logging()



import json
import os
import re
from typing import Any, Dict, List, cast

from chaoslib.exceptions import ActivityFailed
from chaoslib.types import Configuration, Secrets
from slo_generator.compute import compute

__all__ = ["compute_slo"]


def compute_slo(
    slo: Dict[str, Any],
    config: Dict[str, Any],
    configuration: Configuration = None,
    secrets: Secrets = None,
) -> List[Dict[str, Any]]:
    """
    Computes the given SLO and return a list of outcomes for each error budget
    policies in the `config`.

    This is a wrapper around https://github.com/google/slo-generator so all
    of its documentation applies for the definition of the `slo` and
    `config` objects. The former contains the the SLO description while the
    latter describes where to source SLIs from and the error budget policies.

    The most notable difference is that we disable any exporters so there
    is no need to define them in your objects.
    """
    s_config = replace_vars(slo)
    g_config = replace_vars(config)

    # prevent any export
    g_config.pop("default_exporters", None)
    g_config.pop("exporters", None)

    return cast(
        List[Dict[str, Any]],
        compute(
            s_config,
            g_config,
            client=None,
            do_export=False,
        ),
    )


###############################################################################
# Private function
###############################################################################


# drawn from slo-generator to prevent: json -> dict -> yaml -> dict
def replace_vars(
    config: Dict[str, Any], ctx: os._Environ = os.environ  # type: ignore
) -> Dict[str, Any]:
    content = json.dumps(config)

    pattern = re.compile(r".*?\${(\w+)}.*?")
    match = pattern.findall(content)

    if match:
        full_value = content
        for var in match:
            try:
                full_value = full_value.replace(f"${{{var}}}", ctx[var])
            except KeyError:
                raise ActivityFailed(
                    f'Environment variable "{var}" should be set.',
                )

        content = full_value

    return cast(Dict[str, Any], json.loads(content))



from datetime import datetime
from typing import Any, Dict, List, Optional

from chaoslib.exceptions import InvalidActivity
from logzero import logger

from chaosreliably import parse_duration

__all__ = [
    "expire_in_more_than",
    "has_subject_alt_names",
    "has_fingerprint",
    "is_issued_by",
    "verify_tls_cert",
]


def expire_in_more_than(
    duration: str = "7d", value: Optional[Dict[str, Any]] = None
) -> bool:
    """
    Verifies that the certificate expires in more than the given duration.

    The `duration` is expressed as followed: <NUMBER><UNIT> where
    <UNIT> is one of `"s"`, `"m"`, `"d"` or `"w"`. For example, in more
    than a week can be expressed as `"7d"` or `"1w"`.
    """
    v = value["cert"]["not_valid_after"]  # type: ignore
    logger.debug(f"Cert expires on {v}")
    delta = parse_duration(duration)
    expiry_date = datetime.fromisoformat(v)
    return (datetime.utcnow() + delta) < expiry_date


def has_subject_alt_names(
    alt_names: List[str],
    strict: bool = True,
    value: Optional[Dict[str, Any]] = None,
) -> bool:
    """
    Validates the certficate covers at least the given list of alternative
    names. If `strict` is set, then the list of exported names must be exactly
    the provided ones.
    """
    subaltnames = value["cert"]["extensions"].get(  # type: ignore
        "subjectAltName"
    )
    if not subaltnames and alt_names:
        logger.debug("Certificate exposes no alternative subject names")
        return False

    exported = set(subaltnames)
    logger.debug(f"Alt names: {exported} / Expected alt names: {alt_names}")
    if strict:
        return subaltnames == alt_names  # type: ignore

    return set(alt_names).issubset(exported)


def has_fingerprint(
    fingerprint: str,
    hash: str = "sha256",
    value: Optional[Dict[str, Any]] = None,
) -> bool:
    """
    Validate the fingerprint of the certificate. The hash is one of
    `"md5"`, `"sha1"` or `"sha256"`.
    """
    if hash not in ("md5", "sha1", "sha256"):
        raise InvalidActivity(
            "invalid `hash` value in the `has_fingerprint` tolerance"
        )

    fp = value["cert"]["fingerprints"].get(hash)  # type: ignore
    return fp == fingerprint  # type: ignore


def is_issued_by(issuer: str, value: Optional[Dict[str, Any]] = None) -> bool:
    """
    Validate the issue of the certificate.
    """
    return value["cert"]["issuer"] == issuer  # type: ignore


def verify_tls_cert(
    expire_after: str = "7d",
    alt_names: Optional[List[str]] = None,
    fingerprint_sha256: Optional[str] = None,
    issuer: Optional[str] = None,
    value: Optional[Dict[str, Any]] = None,
) -> bool:
    """
    Performs a range of checks on the certificate of the remote endpoint:

    * that we are beyond a certain duration of the certificate expiricy date
    * that the certificate exports the right alternative names
    * the fingerprint of the certificate
    * the certificate was issued by the right issuer

    If any of these values is not set (the default), the according
    check is not performed. This doesn't apply to the expiration date which
    is always checked.
    """
    if expire_in_more_than(expire_after, value) is False:
        return False

    if alt_names not in ("", None):  # type: ignore
        if (
            has_subject_alt_names(alt_names, True, value)  # type: ignore
            is False
        ):
            return False

    if fingerprint_sha256:
        if has_fingerprint(fingerprint_sha256, "sha256", value) is False:
            return False

    if issuer:
        if is_issued_by(issuer, value) is False:
            return False

    return True



import os
import secrets
import threading
from typing import Any, Dict, Optional, Type, Union, cast

from chaosaddons.controls import safeguards
from chaoslib.run import EventHandlerRegistry, RunEventHandler
from chaoslib.types import (
    Configuration,
    Experiment,
    Extension,
    Journal,
    Secrets,
)
from logzero import logger

global_lock = threading.Lock()


class ReliablyGuardian(safeguards.Guardian):  # type: ignore
    def _exit(self) -> None:
        safeguards.Guardian._exit(self)


class ReliablySafeguardGuardian:
    def __init__(
        self,
        url: Union[str, Dict[str, str]],
        auth: Optional[Union[str, Dict[str, str]]],
        frequency: Optional[Union[float, Dict[str, str]]],
        guardian_class: Type[safeguards.Guardian] = ReliablyGuardian,
    ) -> None:
        self.guardian = guardian_class()
        self.frequency = frequency

        url = get_value(url)  # type: ignore
        auth = get_value(auth)  # type: ignore

        if frequency is not None:
            frequency = max(float(get_value(frequency)), 0.3)  # type: ignore

        if not url:
            logger.debug("Missing URL for safeguard/precheck call")
            return None

        name = f"precheck-{secrets.token_hex(8)}"
        self.probes = [
            {
                "name": name,
                "type": "probe",
                "tolerance": True,
                "provider": {
                    "type": "python",
                    "module": "chaosreliably.activities.safeguard.probes",
                    "func": "call_endpoint",
                    "arguments": {"url": url, "auth": auth},
                },
            }
        ]

        if frequency:
            logger.debug(f"Will call '{url}' every {frequency}s as a safeguard")
            self.probes[0]["frequency"] = frequency
        else:
            logger.debug(f"Will call '{url}' once as a precheck")

        self.guardian.prepare(self.probes)

    def start(
        self,
        experiment: Experiment,
        configuration: Configuration,
        secrets: Secrets,
    ) -> None:
        self.guardian.run(
            experiment,
            self.probes,
            configuration or {},
            secrets or {},
            None,
        )

    def finish(self, journal: Journal) -> None:
        self.guardian.terminate()

        with global_lock:
            if self.guardian.was_triggered:
                x = journal["experiment"]
                name = "safeguards" if self.frequency else "prechecks"
                integration = get_integration_from_extension(x, name)
                trig_probes = integration.setdefault("triggered_probes", [])
                trig_probes.append(self.guardian.triggered_by_run)


class ReliablySafeguardHandler(RunEventHandler):  # type: ignore
    def __init__(self) -> None:
        RunEventHandler.__init__(self)
        self._lock = threading.Lock()
        self._initialized = False
        self._started = False

        self.guardians = []  # type: ignore

    @property
    def initialized(self) -> bool:
        with self._lock:
            return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        with self._lock:
            self._initialized = value

    def register(self, event_registry: EventHandlerRegistry) -> None:
        with self._lock:
            if not self._initialized:
                event_registry.register(self)
                self._initialized = True

    def start_all(
        self,
        experiment: Experiment,
        configuration: Configuration,
        secrets: Secrets,
    ) -> None:
        if not self.initialized:
            return None

        extension = find_extension_by_name(experiment, "reliably")
        if extension:
            intg = extension.setdefault("integrations", {})
            intg.setdefault("prechecks", [])
            intg.setdefault("safeguards", [])

        with self._lock:
            if self._started:
                return None

            self._started = True

            for g in self.guardians:
                g.start(experiment, configuration, secrets)

    def finish(self, journal: Journal) -> None:
        if not self.initialized:
            return None

        with self._lock:
            for g in self.guardians:
                try:
                    g.finish(journal)
                except Exception:
                    logger.debug("Guardian somehow failed", exc_info=True)

    def add(self, guardian: ReliablySafeguardGuardian) -> None:
        if not self.initialized:
            return None

        with self._lock:
            self.guardians.append(guardian)


proxy = ReliablySafeguardHandler()


def initialize(
    event_registry: EventHandlerRegistry,
    handler: Optional[ReliablySafeguardHandler] = None,
) -> None:
    (handler or proxy).register(event_registry)


def register(
    url: Union[str, Dict[str, str]],
    auth: Optional[Union[str, Dict[str, str]]] = None,
    frequency: Optional[Union[float, Dict[str, str]]] = None,
    handler: Optional[ReliablySafeguardHandler] = None,
    guardian_class: Type[safeguards.Guardian] = ReliablyGuardian,
) -> None:
    (handler or proxy).add(
        ReliablySafeguardGuardian(
            url, auth, frequency, guardian_class=guardian_class
        )
    )


def run_all(
    experiment: Experiment,
    configuration: Configuration,
    secrets: Secrets,
    handler: Optional[ReliablySafeguardHandler] = None,
) -> None:
    (handler or proxy).start_all(experiment, configuration, secrets)


def find_extension_by_name(
    experiment: Experiment,
    name: str,
) -> Optional[Extension]:
    extensions = experiment.get("extensions")
    if not extensions:
        return None

    for extension in extensions:
        if extension["name"] == name:
            return cast(Extension, extension)

    return None


def get_value(value: Union[str, Dict[str, str]]) -> Optional[str]:
    if not value:
        return None

    if isinstance(value, str):
        return value

    if isinstance(value, float):
        return str(value)

    if value.get("type") == "env":
        return os.getenv(value["key"])

    return None


def get_integration_from_extension(
    experiment: Experiment, name: str
) -> Dict[str, Any]:
    extensions = experiment.setdefault("extensions", [])
    extension = {"name": "reliably"}
    for x in extensions:
        if x["name"] == name:
            extension = x
            break
    else:
        extensions.append(extension)

    integrations = extension.setdefault("integrations", {})  # type: ignore
    integration = integrations.setdefault(name, {})  # type: ignore
    return integration  # type: ignore
